const std = @import("std");
const types = @import("../types.zig");

/// Converts a list of tokens into a list of args decoded to proper
/// union types.
///
/// Caller responsible for freeing returned memory
pub fn argize(allocator: std.mem.Allocator, tokens: []const []const u8) ![]types.MepoArg {
    var args = std.ArrayList(types.MepoArg).init(allocator);
    var i: usize = 0;
    var bracket_start_i: usize = undefined;
    var bracket_nest_n: usize = 0;

    for (tokens) |tok| {
        if (tok[0] == '[' and tok.len == 1) bracket_nest_n += 1;
        if (tok[0] == ']' and tok.len == 1) bracket_nest_n -= 1;

        if (bracket_nest_n == 1 and tok[0] == '[' and tok.len == 1) {
            bracket_start_i = i;
        } else if (bracket_nest_n == 0 and tok[0] == ']' and tok.len == 1) {
            try args.append(.{ .Text = try std.mem.joinZ(allocator, " ", tokens[bracket_start_i + 1 .. i]) });
        } else if (bracket_nest_n >= 1) {
            // nop
        } else if ((tok[0] == '-' and tok.len > 1) or (tok[0] >= '0' and tok[0] <= '9')) {
            try args.append(.{ .Number = try std.fmt.parseFloat(f64, tok) });
        } else {
            try args.append(.{ .Text = try allocator.dupeZ(u8, tok) });
        }
        i += 1;
    }

    return args.items;
}

/// Converts raw text input into tokens, respecting that [, ], and ; should
/// always be treated as seperate tokens
///
/// Caller responsible for freeing returned memory
pub fn tokenize(allocator: std.mem.Allocator, input: []const u8) ![][]const u8 {
    var tokens = std.ArrayList([]const u8).init(allocator);
    var current_token = std.ArrayList(u8).init(allocator);

    for (input) |char| {
        if (char == '[' or char == ']' or char == ';' or char == '\n' or char == ' ') {
            if (current_token.items.len > 0) {
                try tokens.append(current_token.items);
                current_token = std.ArrayList(u8).init(allocator);
            }
            if (char == '[' or char == ']' or char == ';') {
                const char_str: []const u8 = try std.fmt.allocPrint(allocator, "{c}", .{char});
                try tokens.append(char_str);
            }
        } else {
            try current_token.append(char);
        }
    }

    if (current_token.items.len > 0) try tokens.append(current_token.items);
    return tokens.items;
}

/// Test that the given tokens have balanced [] brackets
pub fn brackets_balanced(tokens: []const []const u8) bool {
    var nest: usize = 0;
    for (tokens) |token| {
        if (token.len == 1 and token[0] == '[') nest += 1;
        if (token.len == 1 and token[0] == ']' and nest == 0) return false;
        if (token.len == 1 and token[0] == ']') nest -= 1;
    }
    return nest == 0;
}

/// Converts tokens (generated by tokenize) into seperate statments based on
/// the placement of semicolons (;). Nested [] are ignored so in input like
/// "foo [bar [ bil; jil ]; ]; bar; barn" that would be treated as 3 statements.
///
/// Caller responsible for freeing returned memory
pub fn statementize(allocator: std.mem.Allocator, tokens: []const []const u8) ![][][]const u8 {
    if (!brackets_balanced(tokens)) return error.UnbalancedBrackets;

    var statements = std.ArrayList([][]const u8).init(allocator);
    var current_statement = std.ArrayList([]const u8).init(allocator);
    var nest_level: usize = 0;

    for (tokens) |token| {
        if (token.len == 1 and token[0] == '[') nest_level += 1;
        if (token.len == 1 and token[0] == ']') nest_level -= 1;
        if (nest_level == 0 and token.len == 1 and token[0] == ';') {
            if (current_statement.items.len > 0) try statements.append(current_statement.items);
            current_statement = std.ArrayList([]const u8).init(allocator);
        } else {
            try current_statement.append(token);
        }
    }
    if (current_statement.items.len > 0) try statements.append(current_statement.items);
    return statements.items;
}

test "brackets_balanced" {
    {
        const input = &[_][]const u8{ "foo", "bar", "baz" };
        try std.testing.expect(brackets_balanced(input[0..]));
    }
    {
        const input = &[_][]const u8{ "[", "bar", "]" };
        try std.testing.expect(brackets_balanced(input[0..]));
    }
    {
        const input = &[_][]const u8{ "[", "[", "bar", "]", "[", "bill", "]", "]" };
        try std.testing.expect(brackets_balanced(input[0..]));
    }
    {
        const input = &[_][]const u8{ "[", "bar" };
        try std.testing.expect(!brackets_balanced(input[0..]));
    }
    {
        const input = &[_][]const u8{ "]", "bar", "[" };
        try std.testing.expect(!brackets_balanced(input[0..]));
    }
    {
        const input = &[_][]const u8{ "[", "[", "bar", "]" };
        try std.testing.expect(!brackets_balanced(input[0..]));
    }
}

test "statementize" {
    // Single statement & works without trailing ;
    {
        const input = &[_][]const u8{ "foo", "bar", "baz" };
        const expect_0 = [_][]const u8{ "foo", "bar", "baz" };
        var result = try statementize(std.heap.c_allocator, input[0..]);
        try std.testing.expectEqual(@as(usize, @intCast(1)), result.len);
        for (result[0], 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect_0[idx], result[0][idx]));
        }
    }

    // Multi-statement
    {
        const input = &[_][]const u8{ "foo", "bar", "baz", ";", "bil", "nil", ";" };
        const expect_0 = [_][]const u8{ "foo", "bar", "baz" };
        const expect_1 = [_][]const u8{ "bil", "nil" };
        var result = try statementize(std.heap.c_allocator, input[0..]);
        try std.testing.expectEqual(@as(usize, @intCast(2)), result.len);
        for (result[0], 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect_0[idx], result[0][idx]));
        }
        for (result[1], 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect_1[idx], result[1][idx]));
        }
    }

    // Respects nested []
    {
        const input = &[_][]const u8{ "foo", "[", "bar", "[", "gill", "]", "]", "baz", ";", "bil", "nil", ";" };
        const expect_0 = [_][]const u8{ "foo", "[", "bar", "[", "gill", "]", "]", "baz" };
        const expect_1 = [_][]const u8{ "bil", "nil" };
        var result = try statementize(std.heap.c_allocator, input[0..]);
        try std.testing.expectEqual(@as(usize, @intCast(2)), result.len);
        for (result[0], 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect_0[idx], result[0][idx]));
        }
        for (result[1], 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect_1[idx], result[1][idx]));
        }
    }

    // Returns error for unmatched [] inputs
    {
        const input = &[_][]const u8{ "foo", "]", ";" };
        var result = statementize(std.heap.c_allocator, input[0..]);
        try std.testing.expectEqual(result, error.UnbalancedBrackets);
    }
}

test "tokenize" {
    {
        // basic tokenization test based on whitespace
        const input = "foo bar baz";
        var expect = [_][]const u8{ "foo", "bar", "baz" };
        var result = try tokenize(std.heap.c_allocator, input);
        try std.testing.expectEqual(@as(usize, @intCast(3)), result.len);
        for (result, 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect[idx], result[idx]));
        }
    }

    {
        // e.g. treats [] brackets as individual tokens irregardless of whitespace (test [foo])
        const input = "foo bar [baz]";
        var expect = [_][]const u8{ "foo", "bar", "[", "baz", "]" };
        var result = try tokenize(std.heap.c_allocator, input);
        try std.testing.expectEqual(@as(usize, @intCast(5)), result.len);
        for (result, 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect[idx], result[idx]));
        }
    }

    {
        // e.g. treats [] brackets as individual tokens irregardless of whitespace (test [ foo])
        const input = "foo bar [ baz]";
        var expect = [_][]const u8{ "foo", "bar", "[", "baz", "]" };
        var result = try tokenize(std.heap.c_allocator, input);
        try std.testing.expectEqual(@as(usize, @intCast(5)), result.len);
        for (result, 0..) |_, idx| {
            try std.testing.expect(std.mem.eql(u8, expect[idx], result[idx]));
        }
    }
}

test "argize" {
    var arena = std.heap.ArenaAllocator.init(std.testing.allocator);
    defer arena.deinit();

    const spec = struct {
        input: []const []const u8,
        expect: []const types.MepoArg,
    };

    var specs = [_]spec{
        .{
            // Standard check to make sure text / numbers argized properly
            .input = &[_][]const u8{ "arga", "22.2" },
            .expect = &[_]types.MepoArg{
                types.MepoArg{ .Text = "arga" },
                types.MepoArg{ .Number = 22.2 },
            },
        },
        .{
            // Numbers check (negative, postive)
            .input = &[_][]const u8{ "-22.4", "2747.2716" },
            .expect = &[_]types.MepoArg{
                types.MepoArg{ .Number = -22.4 },
                types.MepoArg{
                    .Number = 2747.2716,
                },
            },
        },
        .{
            // Ensure argizer never parses past 1st set of string, e.g. nested []
            // should just tokenize as single arg proc
            .input = &[_][]const u8{ "foo", "bar", "[", "baz", "bill", "]", "dill", "jill" },
            .expect = &[_]types.MepoArg{ types.MepoArg{ .Text = "foo" }, types.MepoArg{ .Text = "bar" }, types.MepoArg{ .Text = "baz bill" }, types.MepoArg{ .Text = "dill" } },
        },
    };

    for (specs) |s| {
        const result = try argize(arena.allocator(), s.input[0..]);
        for (s.expect, 0..) |_, i| {
            try std.testing.expectEqual(@TypeOf(s.expect[i]), @TypeOf(result[i]));
            switch (result[i]) {
                .Number => try std.testing.expectEqual(s.expect[i].Number, result[i].Number),
                .Text => try std.testing.expectEqualSlices(u8, s.expect[i].Text, result[i].Text),
            }
        }
    }
}
